//
// Example of watercolor rendering via GLSL programs
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#include    <glut.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "TypeDefs.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "GlslProgram.h"

Vector3D    eye   ( -0.5, -0.5, 1.5 );  // camera position
Vector3D    light ( 5, 0, 4 );          // light position
unsigned	decalMap;                   // decal (diffuse) texture
unsigned   	stoneMap;
unsigned   	teapotMap;
unsigned   	noiseMap;
unsigned    paperMap;
Vector3D    rot ( 0, 0, 0 );
float       angle     = 0;
int         mouseOldX = 0;
int         mouseOldY = 0;
bool        useFilter = true;

GlslProgram program;

void    startOrtho ()
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, 512, 0, 512, -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix
}

void init ()
{
    glClearColor ( 1.0, 1.0, 1.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
    glEnable     ( GL_TEXTURE_2D );
    glDepthFunc  ( GL_LEQUAL     );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void display ()
{
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    startOrtho ();

    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_2D, paperMap );
    glEnable           ( GL_TEXTURE_2D );

    glActiveTextureARB ( GL_TEXTURE1_ARB );
    glDisable          ( GL_TEXTURE_2D );

    glActiveTextureARB ( GL_TEXTURE2_ARB );
    glDisable          ( GL_TEXTURE_2D );

    glDepthMask        ( GL_FALSE );
    glColor4f          ( 1, 1, 1, 1 );
/*
    glBegin ( GL_QUADS );

        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( 1,   0 );
        glVertex2f   ( 511, 0 );

        glTexCoord2f ( 1, 1 );
        glVertex2f   ( 511, 511 );

        glTexCoord2f ( 0, 1 );
        glVertex2f   ( 0, 511 );
    glEnd   ();
*/
    endOrtho ();


    glDepthMask        ( GL_TRUE );

    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_2D, teapotMap );

    glActiveTextureARB ( GL_TEXTURE1_ARB );
    glBindTexture      ( GL_TEXTURE_2D, noiseMap );

    glActiveTextureARB ( GL_TEXTURE2_ARB );
    glBindTexture      ( GL_TEXTURE_2D, paperMap );

    glActiveTextureARB ( GL_TEXTURE0_ARB );

    if ( useFilter )
        program.bind ();

    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );

    glutSolidTeapot ( 0.4 );

    glPopMatrix ();

    if ( useFilter )
        program.unbind ();

    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,    // eye
                    0, 0, 0,                // center
                    0, 0, 1 );              // up
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );

    if ( key == 'f' || key == 'F' )
        useFilter = !useFilter;
}

void    animate ()
{
    angle  = 0.001f * glutGet ( GLUT_ELAPSED_TIME );

    light.x = 2*cos ( angle );
    light.y = 2*sin ( angle );
    light.z = 3 + 0.3 * sin ( angle / 3 );

    program.bind ();
    program.setUniformVector ( "eyePos",   eye   );
    program.setUniformVector ( "lightPos", light );
    program.setUniformFloat  ( "time",     angle );
    program.unbind ();

    glutPostRedisplay ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
        rot.z -= 360;

    if ( rot.z < -360 )
        rot.z += 360;

    if ( rot.y > 360 )
        rot.y -= 360;

    if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

    glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
    }
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    glutInitWindowSize  ( 512, 512 );


                                // create window
    glutCreateWindow ( "OpenGL watercolor rendering with edge softening" );

                                // register handlers
    glutDisplayFunc  ( display );
    glutReshapeFunc  ( reshape );
    glutKeyboardFunc ( key     );
    glutMouseFunc    ( mouse   );
    glutMotionFunc   ( motion  );
    glutIdleFunc     ( animate );

    init           ();
    initExtensions ();

	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported.\n" );

		return 1;
	}

    if ( !program.loadShaders ( "watercolor-2.vsh", "watercolor-2.fsh" ) )
    {
        printf ( "Error loading shaders:\n%s\n", program.getLog ().c_str () );

        return 3;
    }

    decalMap  = createTexture2D ( true, "../../Textures/oak.bmp" );
    stoneMap  = createTexture2D ( true, "../../Textures/block.bmp" );
    teapotMap = createTexture2D ( true, "../../Textures/Oxidated.jpg" );

    noiseMap = createTexture2D ( true, "noise-2D.png" );
    paperMap  = createTexture2D ( true, "paper.dds"  );

    program.bind ();
    program.setTexture ( "noiseMap", 1 );
    program.setTexture ( "paperMap", 2 );
    program.unbind ();

//    printf ( "Render scene in watercolor mode\n" );
//    printf ( "Press F key to turn watercolor mode on/off\n" );

    glutMainLoop ();

    return 0;
}
